///////////////////////////////////////////////////////////////////////
//if COORD_ALGO = 0 then:
///////////////////////////////////////////////////////////////////////
//a demining scenario with a single robot moving in a zig zag
//pattern. the robot stops if : (a) it explodes; or (b) it has exceeded
//its deadline; or (c) if it has covered 1/N of the total number of
//cells. in the first two cases, the robot has failed. in the last case,
//the robot has succeeded. we compute the probability of success and
//raise it to the N-th power. this way, we get the probability of
//success if N robots independently demined 1/N of the field. varying N
//gives us different results, and we want the smallest N for which the
//result exceeds 0.8.
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
//if COORD_ALGO = 1 then:
///////////////////////////////////////////////////////////////////////
//a demining scenario with a team of N robots moving in a zig zag
//pattern. the team consists of a leader, and followers maintaining a
//fixed distance from the leader. if the leader explodes, then the
//followers (if any left) elect a new leader. the team stops if : (a)
//the leader explodes and no robots are left; or (b) it has exceeded its
//deadline; or (c) if it has covered all cells. in the first two cases,
//the team has failed. in the last case, the team has succeeded. we
//compute the probability of success. varying N gives us different
//results, and we want the smallest N for which the result exceeds 0.8.
///////////////////////////////////////////////////////////////////////

dtmc

///////////////////////////////////////////////////////////////////////
//parameters supplied by command line
///////////////////////////////////////////////////////////////////////
//number of robots
const int NUM_ROBOTS;
//detecting capability -- should be 0, 1 or 2
const int DETECTING_CAPABILITY;
//defusing capability -- should be 0, 1 or 2
const int DEFUSING_CAPABILITY;
//the coordination algorithm -- should be 0 or 1
const int COORD_ALGO;

//the terrain -- maps detecting capability to probability of a false
//negative, and gives probabilities for being in the three defusing
//situations.
const double P_FN_SC0;
const double P_FN_SC1;
const double P_FN_SC2;
const double P_IN_DEFUSE1;
const double P_IN_DEFUSE2;
const double P_IN_DEFUSE3;

///////////////////////////////////////////////////////////////////////
//constants
///////////////////////////////////////////////////////////////////////

//robot states
const int INIT = 0;
const int NEXT = 1;
const int WORK = 2;
const int BLOWNUP = 3;
const int DONE = 4;
const int NOT_DETECTED_MINE = 5;
const int DEFUSE1 = 6;
const int DEFUSE2 = 7;
const int DEFUSE3 = 8;
const int DEFUSE = 9;
const int MARK = 10;
const int TO_EXPLODE = 11;
const int TIMEOUT = 12;
const int SENSE_MINE = 13;
const int ELECT_LEADER = 14;
const int LAST = 14;

//maximum X and Y axis values -- smallest is 0
const int X_MAX = 9;
const int Y_MAX = 9;

//probability of blowing up while detecting a mine -- this depends on
//the detecting capability
const double p_explode_detect = (DETECTING_CAPABILITY=0) ? 0.0001 :
                                (DETECTING_CAPABILITY=1) ? 0.00001 : 0.000001;

//probability of detecting a mine. we assume no false positives. so if
//a mine is detected, then there is a mine in the cell.
const double p_detect_mine = 0.5;

//probabilty of false negative -- this depends on the detecting capability
const double p_false_neg = (DETECTING_CAPABILITY=0) ? P_FN_SC0 : 
                           (DETECTING_CAPABILITY=1) ? P_FN_SC1 : P_FN_SC2;

//probabilities of blowing up in each defusing state. this depends on
//the defusing capability.
const double p_explode_defuse1 = (DEFUSING_CAPABILITY=0) ? 0.01 : 
                                 (DEFUSING_CAPABILITY=1) ? 0.001 : 0.0001;
const double p_explode_defuse2 = (DEFUSING_CAPABILITY=0) ? 0.01 : 
                                 (DEFUSING_CAPABILITY=1) ? 0.001 : 0.0001;
const double p_explode_defuse3 = (DEFUSING_CAPABILITY=0) ? 0.01 : 
                                 (DEFUSING_CAPABILITY=1) ? 0.001 : 0.0001;
const double p_explode_mark = (DEFUSING_CAPABILITY=0) ? 0.01 : 
                              (DEFUSING_CAPABILITY=1) ? 0.001 : 0.0001;

//probability of success in leader election
const double p_elect_leader = 0.999;

//various times
const int t_sense = 1; //time to sense mine
const int t_defuse1 = 1; //time to defuse in easy case
const int t_defuse2 = 2; //time to defuse in medium case
const int t_defuse3 = 3; //time to defuse in hard case
const int t_mark = 4; //time to defuse in hard case
const int t_elect_leader = 1; //time to elect new leader

//initial coordinates of robot
const int INIT_X = 0;
const int INIT_Y = 0;

//the number of cells to be covered
const int NUM_CELLS = 
      (COORD_ALGO=0) ? ceil((X_MAX+1) * (Y_MAX+1) / NUM_ROBOTS) :
      ((X_MAX+1) * (Y_MAX+1));

//the deadline
const int DEADLINE = 400;

///////////////////////////////////////////////////////////////////////
//robot 1
///////////////////////////////////////////////////////////////////////
module Robot1
    //local state
    s1 : [INIT..LAST] init INIT;

    //whether all operations succeeded
    success1 : bool init true;

    //current X and Y coordinate value
    x1 : [0..X_MAX] init INIT_X;
    y1 : [0..Y_MAX] init INIT_Y;
    dir1 : bool init false;

    //clock, i.e., time spent so far
    clock1 : [0..DEADLINE] init 0;

    //number of cells covered so far
    cells1 : [0..NUM_CELLS] init 0;

    //team size
    team_size1 : [0..NUM_ROBOTS] init NUM_ROBOTS;

    //go to work
    [] s1=INIT -> (s1'=WORK);

    //if all cells covered, we are done
    [] (s1=WORK)&(cells1=NUM_CELLS) -> (s1'=DONE);

    //if not all cells covered, increment the number of cells
    //covered and continue to checking for timeout. 
    [] (s1=WORK)&(cells1!=NUM_CELLS) -> (s1'=SENSE_MINE)&(cells1'=cells1+1);

    //stutter in timeout state
    [] s1=TIMEOUT -> (s1'=TIMEOUT);

    //timeout can happen during mine detecting
    [] (s1=SENSE_MINE)&(clock1+t_sense>DEADLINE) -> (s1'=TIMEOUT); 

    //after detecting, if no timeout happends, either a mine was
    //detected or not, or the robot blew up. also, we increase the
    //clock by the time it took to sense.
    [] (s1=SENSE_MINE)&(clock1+t_sense<=DEADLINE) -> 
       p_explode_detect:(s1'=TO_EXPLODE)&(clock1'=clock1+t_sense) + 
       p_detect_mine:(s1'=DEFUSE1)&(clock1'=clock1+t_sense) + 
       (1-p_explode_detect-p_detect_mine):(s1'=NOT_DETECTED_MINE)&(clock1'=clock1+t_sense);

    //if mine not detected, blow up or move to next (successfully,
    //i.e., there really was no mine, or with failure, i.e., a mine
    //was there but was not detected). we assume that if there was a
    //mine (i.e., a false negative), then there is a .5 probability of
    //blowing up.
    [] s1=NOT_DETECTED_MINE -> p_false_neg * 0.5:(s1'=TO_EXPLODE) + 
                               p_false_neg * 0.5:(s1'=NEXT)&(success1'=false) +
                               (1-p_false_neg):(s1'=NEXT);

    //timeout can happen during defuse1
    [] (s1=DEFUSE1)&(clock1+t_defuse1>DEADLINE) -> (s1'=TIMEOUT);

    //if defuse1, next, defuse2, or explode
    [] (s1=DEFUSE1)&(clock1+t_defuse1<=DEADLINE) -> 
       (P_IN_DEFUSE1*(1-p_explode_defuse1)):(s1'=DEFUSE)&(clock1'=clock1+t_defuse1) + 
       P_IN_DEFUSE1*p_explode_defuse1:(s1'=TO_EXPLODE)&(clock1'=clock1+t_defuse1) + 
       (1-P_IN_DEFUSE1):(s1'=DEFUSE2);

    //timeout can happen during defuse2
    [] (s1=DEFUSE2)&(clock1+t_defuse2>DEADLINE) -> (s1'=TIMEOUT);

    //if defuse2, next, defuse3, or explode
    [] (s1=DEFUSE2)&(clock1+t_defuse2<=DEADLINE) -> 
       (P_IN_DEFUSE2*(1-p_explode_defuse2)):(s1'=DEFUSE)&(clock1'=clock1+t_defuse2) + 
       P_IN_DEFUSE2*p_explode_defuse2:(s1'=TO_EXPLODE)&(clock1'=clock1+t_defuse2) + 
       (1-P_IN_DEFUSE2):(s1'=DEFUSE3);

    //timeout can happen during defuse3
    [] (s1=DEFUSE3)&(clock1+t_defuse3>DEADLINE) -> (s1'=TIMEOUT);

    //if defuse3, next, mark, or explode
    [] (s1=DEFUSE3)&(clock1+t_defuse3<=DEADLINE) -> 
       (P_IN_DEFUSE3*(1-p_explode_defuse3)):(s1'=DEFUSE)&(clock1'=clock1+t_defuse3) + 
       P_IN_DEFUSE3*p_explode_defuse3:(s1'=TO_EXPLODE)&(clock1'=clock1+t_defuse3) + 
       (1-P_IN_DEFUSE3):(s1'=MARK);

    //if defuse, next
    [] s1=DEFUSE -> (s1'=NEXT);

    //timeout can happen during mark
    [] (s1=MARK)&(clock1+t_mark>DEADLINE) -> (s1'=TIMEOUT);

    //if mark, next, or explode
    [] (s1=MARK)&(clock1+t_mark<=DEADLINE) -> 
       (1-p_explode_mark):(s1'=NEXT)&(clock1'=clock1+t_mark) + 
       p_explode_mark:(s1'=TO_EXPLODE)&(clock1'=clock1+t_mark);

    //if to explode, and COORD_ALGO=0, blow up
    [] (s1=TO_EXPLODE)&(COORD_ALGO=0) -> (s1'=BLOWNUP);

    //if to explode, and COORD_ALGO=1, try to elect new leader
    [] (s1=TO_EXPLODE)&(COORD_ALGO=1) -> (s1'=ELECT_LEADER);

    //leader election: no more robots left in team, blow up
    [] (s1=ELECT_LEADER)&(team_size1=1) -> (s1'=BLOWNUP);

    //leader election: more robots left, but timeout can happen
    [] (s1=ELECT_LEADER)&(team_size1>1)&(clock1+t_elect_leader>DEADLINE) -> (s1'=TIMEOUT);

    //leader election: if more robots left in team, probabilistically elect new leader and resume
    [] (s1=ELECT_LEADER)&(team_size1>1)&(clock1+t_elect_leader<=DEADLINE) -> 
       p_elect_leader:(s1'=NEXT)&(team_size1'=team_size1-1)&(clock1'=clock1+t_elect_leader) +
       (1-p_elect_leader):(s1'=BLOWNUP)&(team_size1'=team_size1-1)&(clock1'=clock1+t_elect_leader);

    //if blown up, be stuck
    [] s1=BLOWNUP -> (s1'=BLOWNUP);

    //choosing the next cell

    //if you are at the top moving up, move right and then down
    [right1] (s1=NEXT)&(!dir1)&(y1=Y_MAX)&(x1+1<=X_MAX) -> (s1'=WORK)&(dir1'=true)&(x1'=x1+1);

    //if you are at the top moving up, and cannot move right, then you are done
    [] (s1=NEXT)&(!dir1)&(y1=Y_MAX)&(x1+1>X_MAX) -> (s1'=DONE);

    //if you are not at the top moving up, move up
    [up1] (s1=NEXT)&(!dir1)&(y1<Y_MAX) -> (s1'=WORK)&(y1'=y1+1);

    //if you are at the bottom moving down, move right and then up
    [right1] (s1=NEXT)&(dir1)&(y1=0)&(x1+1<=X_MAX) -> (s1'=WORK)&(dir1'=false)&(x1'=x1+1);

    //if you are at the bottom moving down, and cannot move right, then you are done
    [] (s1=NEXT)&(dir1)&(y1=0)&(x1+1>X_MAX) -> (s1'=DONE);

    //if you are not at the bottom moving down, move down
    [down1] (s1=NEXT)&(dir1)&(y1>0) -> (s1'=WORK)&(y1'=y1-1);

    //if done, be stuck
    [] s1=DONE -> (s1'=DONE);
endmodule

///////////////////////////////////////////////////////////////////////
//reward structure to keep track of number of cells covered by robot1
///////////////////////////////////////////////////////////////////////
rewards "covered_cells1"
    [] (s1=SENSE_MINE) : 1.0;
endrewards

///////////////////////////////////////////////////////////////////////
//reward structure to keep track of number of cells marked by robot1
///////////////////////////////////////////////////////////////////////
rewards "marked_cells1"
    [] (s1=MARK) : 1.0;
endrewards

///////////////////////////////////////////////////////////////////////
//reward structure to keep track of number of cells defused by robot1
///////////////////////////////////////////////////////////////////////
rewards "defused_cells1"
    [] (s1=DEFUSE) : 1.0;
endrewards

